/*
************************************************************************
**	O.S.		: Linux
**	FILE NAME	: osm.h
**	Author		: C.L. Huang   
**	Description	: SCSI Device Driver for Areca Host Bus Adapter 
************************************************************************
** Copyright (C) 2002 - 2012, Areca Technology Corporation All rights reserved.
**
** Web site: www.areca.com.tw
** E-mail: support@areca.com.tw
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
** This program is distributed in the hope that it will be useful, but 
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
************************************************************************
** Redistribution and use in source and binary forms,with or without
** modification,are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice,this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice,this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
** IMPLIED WARRANTIES,INCLUDING,BUT NOT LIMITED TO,THE IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
** IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,INDIRECT,
** INCIDENTAL,SPECIAL,EXEMPLARY,OR CONSEQUENTIAL DAMAGES(INCLUDING,BUT
** NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA,OR PROFITS; OR BUSINESS INTERRUPTION)HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY,WHETHER IN CONTRACT,STRICT LIABILITY,OR TORT
**(INCLUDING NEGLIGENCE OR OTHERWISE)ARISING IN ANY WAY OUT OF THE USE OF
** THIS SOFTWARE,EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
************************************************************************
*/
#if defined __KERNEL__
	#include <linux/version.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
	#define MODVERSIONS
#endif
	/*modversions.h should be before should be before module.h*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#if defined(MODVERSIONS)
	#include <config/modversions.h>
#endif
#endif
	#include <linux/kernel.h>
	#include <linux/types.h>
	#include <scsi/scsi_host.h>
	#include <linux/pci.h>
	#include <linux/device.h>
	#include <linux/dmapool.h>
	#include <linux/dma-mapping.h>
	#include <linux/sched.h>
	#include <linux/jiffies.h>
	#include <linux/module.h>
	#include <asm/dma.h>
	#include <asm/io.h>
	#include <asm/uaccess.h>
	#include <linux/delay.h>
	#include <linux/signal.h>
	#include <linux/errno.h>
	#include <linux/ioport.h>
	#include <linux/proc_fs.h>
	#include <linux/string.h>
	#include <linux/ctype.h>
	#include <linux/interrupt.h>
	#include <linux/stddef.h>
	#include <linux/pci_ids.h>
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)) && (LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 0)))
#ifndef PCIE_LINK_STATE_L0S
	#include <linux/pci-aspm.h>
#endif
#endif
	#include <linux/moduleparam.h>
	#include <linux/blkdev.h>
	#include <linux/timer.h>
	#include <linux/reboot.h>
	#include <linux/sched.h>
	#include <linux/init.h>
	#include <linux/highmem.h>
	#include <linux/spinlock.h>
	#include <linux/nmi.h>
	#include <linux/kthread.h>
	#include <linux/rtc.h>
//	#include <asm/rtc.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
	#include <linux/pci_regs.h>
	#include <linux/sched.h>
#endif
	#include <scsi/scsi.h>
	#include <scsi/scsi_device.h>
	#include <scsi/scsi_cmnd.h>
	#include <scsi/scsi_tcq.h>
	#include <scsi/scsicam.h>
	#include <scsi/scsi_eh.h>
	#include <scsi/scsi_transport.h>
	#include <linux/string.h>
	#include <linux/workqueue.h>
	#include <linux/hardirq.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 9)
	#include <linux/transport_class.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 2, 0)
	#include <linux/vmalloc.h>
#endif
#endif

#define ARCSAS_DBG_SUSPECT			0
#define ARCSAS_DBG_FUNCTION			0
#define ARCSAS_DBG_CONTEXT			0
#define ARCSAS_DBG_SGIO				0
#define ARCSAS_DBG_SGIO_CONTEXT			0
#define ARCSAS_DBG_INIT				0
#define ARCSAS_DBG_INIT_CONTEXT			0
#define ARCSAS_DBG_LINKEDLIST			0
#define ARCSAS_DBG_UPDATE_FLASH_DEVICE_MAP	0
#define ARCSAS_DBG_OS_CCB_CONTEXT		0
#define ARCSAS_DBG_OS_CCB			0
#define ARCSAS_DBG_INTERRUPT			0
#define ARCSAS_DBG_DISCOVERY			0
#define ARCSAS_DBG_DISCOVERY_CONTEXT		0
#define ARCSAS_DBG_SATA_SM			0
#define ARCSAS_DBG_SATA_SM_CONTEXT		0
#define ARCSAS_DBG_SAS_SM			0
#define ARCSAS_DBG_SAS_SM_CONTEXT		0
#define ARCSAS_DBG_PM_SM			0
#define ARCSAS_DBG_PM_SM_CONTEXT		0
#define ARCSAS_DBG_PORT_SM			0
#define ARCSAS_DBG_PORT_SM_CONTEXT		0
#define ARCSAS_DBG_SORT_INDEX			0
#define ARCSAS_DBG_SORT_INDEX_CONTEXT		0
#define ARCSAS_DBG_SES				0
#define ARCSAS_DBG_SES_CONTEXT			0
#define ARCSAS_DBG_IOCTL			0
#define ARCSAS_DBG_IOCTL_CONTEXT		0
#define ARCSAS_DBG_IOCTL_SET_DISK_CONFIG	0
#define ARCSAS_DBG_IOCTL_GET_DISK_CONFIG	0
#define ARCSAS_DBG_IOCTL_SET_ADPATER_CONFIG	0
#define ARCSAS_DBG_IOCTL_GET_ADPATER_CONFIG	0
#define ARCSAS_DBG_IOCTL_EVENT			0
#define ARCSAS_DBG_IOCTL_GET_SMART_STATUS	0
#define ARCSAS_DBG_IOCTL_PASS_THROU		0
#define ARCSAS_DBG_VIRTUAL_SES_PASS_THROU	0
#define ARCSAS_DBG_SGPIO			0
#define ARCSAS_DBG_ATA_PASSTHROU		0
#define ARCSAS_DBG_ATA_PASSTHROU_CONTEXT	0
#define ARCSAS_DBG_SPI				0
#define ARCSAS_DBG_SPI_CONTEXT			0
#define ARCSAS_DBG_HIBERNATE			0
#define ARCSAS_DBG_INFO				0
#define ARCSAS_DBG_REGS				0
#define ARCSAS_DBG_CDB				0

#define ARCSAS_DBG_TIMEOUT			0
#define ARCSAS_DBG_TASKFILE_ERR			0
#define ARCSAS_DBG_ERRINFO			0
#define ARCSAS_DBG_OS_EH			0
#define ARCSAS_DBG_PCI_REMOVE			0
#define ARCSAS_DBG_PLUG_UNPLUG			0
#define ARCSAS_DBG_PORT_ABORT_CCB		0
#define ARCSAS_DBG_SLOT_RESET			0
#define ARCSAS_DBG_CHECK_CONDITION		0
#define ARCSAS_DBG_HAS_SENSE_ERROR		0
#define ARCSAS_DBG_CMD_SUPPORT			0

#define ARCSAS_DBG_MARVELL			0

#define ARCSAS_SUPPORT_MSI			1
#define ARCSAS_SUPPORT_MSIX			1
#define ARCSAS_SUPPORT_SES			1
#define ARCSAS_SUPPORT_SGPIO			1
#define ARCSAS_SUPPORT_IOCTL			1
#define ARCSAS_SUPPORT_IOCTL_LOG		0
#define ARCSAS_SUPPORT_SGPIO_TEST		0
#define ARCSAS_SUPPORT_SGPIO_SDATA_IN_OUT	0
#define ARCSAS_SUPPORT_SGPIO_REG_READ_WRITE	0
#define SUPPORT_STAGGERED_SPIN_UP		1
#define SUPPORT_DIRECT_SATA_SSU			1
#define SUPPORT_ACTIVE_CABLE			0
#define SUPPORT_PHY_POWER_MODE			0
#define SUPPORT_I2C				0
#define SUPPORT_I2C_SES				0
#define SAS_PLUGFEST_CABLE			0

extern unsigned char modePageBuf[];
extern unsigned char info_except_ctrl_mode_page[];
extern struct timezone sys_tz;
typedef dma_addr_t osm_dma_addr_t;

#ifndef offsetof
	#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

#ifndef container_of
#define container_of(ptr, type, member) ({          \
	const typeof(((type *)0)->member)*__mptr = (ptr);    \
		     (type *)((char *)__mptr - offsetof(type, member)); })
#endif
/*
************************************************************************
************************************************************************
*/
#define ARCSAS_SWAP_32(x)	(((u32)((unsigned char)(x))) << 24 | ((u32)((unsigned char)((x) >> 8))) << 16 | ((u32)((unsigned char)((x) >> 16))) << 8 | ((u32)((unsigned char)((x) >> 24))))
#define ARCSAS_SWAP_64(x)	((((unsigned long long)) (ARCSAS_SWAP_32((x).parts.low))) << 32 | ARCSAS_SWAP_32((x).parts.high))
#define ARCSAS_SWAP_16(x)	(((u16) ((unsigned char) (x))) << 8 | (u16) ((unsigned char) ((x) >> 8)))
#ifdef _BIG_ENDIAN__
	#define ARCSAS_CPU_TO_LE16_PTR(pu16)	*((unsigned short *)(pu16)) = ARCSAS_CPU_TO_LE16(*(unsigned short *)(pu16))	/*cpu_to_le16p(pu16)*/
	#define ARCSAS_CPU_TO_LE32_PTR(pu32)	*((unsigned int *)(pu32)) = ARCSAS_CPU_TO_LE32(*(unsigned int *)(pu32))		/*cpu_to_le32p(pu32)*/
	#define ARCSAS_LE16_TO_CPU_PTR(pu16)	*((unsigned short *)(pu16)) = ARCSAS_LE16_TO_CPU(*(unsigned short *)(pu16))	/*le16_to_cpup(pu16)*/
	#define ARCSAS_LE32_TO_CPU_PTR(pu32)	*((unsigned int *)(pu32)) = ARCSAS_LE32_TO_CPU(*(unsigned int *)(pu32))		/*le32_to_cpup(pu32)*/
#else
	#define ARCSAS_CPU_TO_LE16_PTR(pu16)        
	#define ARCSAS_CPU_TO_LE32_PTR(pu32)        
	#define ARCSAS_LE16_TO_CPU_PTR(pu32)
	#define ARCSAS_LE32_TO_CPU_PTR(pu32)
#endif
#define ARCSAS_CPU_TO_LE16	cpu_to_le16
#define ARCSAS_CPU_TO_LE32	cpu_to_le32
#define ARCSAS_CPU_TO_LE64(x)	cpu_to_le64((x).value)
#define ARCSAS_CPU_TO_BE16	cpu_to_be16
#define ARCSAS_CPU_TO_BE32	cpu_to_be32
#define ARCSAS_CPU_TO_BE64(x)	cpu_to_be64((x).value)

#define ARCSAS_LE16_TO_CPU	le16_to_cpu
#define ARCSAS_LE32_TO_CPU	le32_to_cpu
#define ARCSAS_LE64_TO_CPU(x)	le64_to_cpu((x).value)
#define ARCSAS_BE16_TO_CPU	be16_to_cpu
#define ARCSAS_BE32_TO_CPU	be32_to_cpu
#define ARCSAS_BE64_TO_CPU(x)	be64_to_cpu((x).value)

/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 32)
#define PCIE_LINK_STATE_L0S	1
#define PCIE_LINK_STATE_L1	2
#define PCIE_LINK_STATE_CLKPM	4
#endif
/*
************************************************************************
************************************************************************
*/
#ifndef SCSI_DEFAULT_MAX_SECTORS
	#define SCSI_DEFAULT_MAX_SECTORS	1024
#endif
/*
************************************************************************
************************************************************************
*/
#ifndef DMA_BIT_MASK
	#define DMA_BIT_MASK(n)	(((n) == 64) ? ~0ULL : ((1ULL<<(n))-1))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 16, 0)
#define pci_pool_create(name, pdev, size, align, allocation) \
		dma_pool_create(name, &pdev->dev, size, align, allocation)
#define	pci_pool_destroy(pool) dma_pool_destroy(pool)
#define	pci_pool_alloc(pool, flags, handle) dma_pool_alloc(pool, flags, handle)
#define	pci_pool_free(pool, vaddr, addr) dma_pool_free(pool, vaddr, addr)
#endif
/*
************************************************************************
**        split 64bits dma addressing
************************************************************************
*/
#define dma_addr_hi32(addr)		(uint32_t) ((addr >> 16) >> 16)
#define dma_addr_lo32(addr) 		(uint32_t) (addr & 0xffffffff)
/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 25)
	#define arcsas_use_sg(cmd)	scsi_sg_count(cmd)
	#define arcsas_rq_bf(cmd)	scsi_sglist(cmd)
	#define arcsas_rq_bflen(cmd)	scsi_bufflen(cmd)
	#define arcsas_map_sg(cmd)	scsi_dma_map(cmd)
	#define arcsas_unmap_sg(cmd)	scsi_dma_unmap(cmd)
	#define arcsas_reassign_bflen(cmd)	cmd->sdb.length
	#define arcsas_scsi_resid(cmd)	cmd->sdb.resid
#else
	#define arcsas_use_sg(cmd)	cmd->use_sg
	#define arcsas_rq_bf(cmd)	cmd->request_buffer
	#define arcsas_rq_bflen(cmd)	cmd->request_bufflen
	#define arcsas_map_sg(cmd)	pci_map_sg(pACB->pcidev, arcsas_rq_bf(cmd), arcsas_use_sg(cmd), cmd->sc_data_direction)
	#define arcsas_unmap_sg(cmd)	pci_unmap_sg(pACB->pcidev, arcsas_rq_bf(cmd), arcsas_use_sg(cmd), cmd->sc_data_direction)
	#define arcsas_reassign_bflen(cmd)	cmd->request_bufflen
	#define arcsas_scsi_resid(cmd)	cmd->resid
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
	#define arcsas_map_sg_page(sg)	kmap_local_page(sg_page(sg))
#elif LINUX_VERSION_CODE > KERNEL_VERSION(3, 3, 0)
	#define arcsas_map_sg_page(sg)	kmap_atomic(sg_page(sg))
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
	#define arcsas_map_sg_page(sg)	kmap_atomic(sg_page(sg), KM_IRQ0)
#else
	#define arcsas_map_sg_page(sg)	kmap_atomic(sg->page, KM_IRQ0)
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
	#define arcsas_unmap_buffer(buffer)	kunmap_local(buffer)
#elif LINUX_VERSION_CODE > KERNEL_VERSION(3, 3, 0)
	#define arcsas_unmap_buffer(buffer)	kunmap_atomic(buffer)
#else
	#define arcsas_unmap_buffer(buffer)	kunmap_atomic(buffer, KM_IRQ0)
#endif

#if !defined(scsi_sglist) && LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
	#define scsi_sglist(cmd) ((struct scatterlist *)(cmd)->request_buffer)
#endif
#ifndef scsi_for_each_sg
	#define scsi_for_each_sg(cmd, sg, nseg, i)			\
	 for (i = 0, sg = scsi_sglist(cmd); i < (nseg); i++, (sg)++)
	static inline struct scatterlist *sg_next(struct scatterlist *sg)
	{
		return sg + 1;
	}
#endif
typedef _Bool	Bool;

#ifndef false
	#define false	0
#endif
#ifndef true
	#define true	1
#endif

#ifndef SCSI_QDEPTH_DEFAULT
#define SCSI_QDEPTH_DEFAULT	0
#endif

#define osm_vmalloc	vmalloc
#define osm_vfree	vfree

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
static inline void scsi_set_resid(struct scsi_cmnd *cmd, int resid)
{
	cmd->resid = resid;
}
#endif

#ifndef PCI_EXP_LNKCTL
#define PCI_EXP_LNKCTL	16	/* Link Control */
#endif

/*
************************************************************************
************************************************************************
*/
#ifdef CONFIG_64BIT
	#define __KCONF_64BIT__
#endif /* CONFIG_64BIT */
/*
************************************************************************
************************************************************************
*/
#if defined(__KCONF_64BIT__)||defined(_64_SYS_)
	#define _SUPPORT_64_BIT
#else
	#ifdef _SUPPORT_64_BIT
		#error Error 64_BIT CPU Macro
	#endif
#endif
#if defined(_SUPPORT_64_BIT)
	typedef unsigned long long	PTR_INTEGER;
#else
	typedef unsigned long	PTR_INTEGER;
#endif
/*
************************************************************************
************************************************************************
*/
#if defined(_SUPPORT_64_BIT)
	#define PtrToNum(p)	(unsigned long long)(void *)(p)
	#define NumToPtr(ul)	(void *)((unsigned long long)ul)
#else
	#define PtrToNum(p)	(unsigned int)(unsigned char *)(p)
	#define NumToPtr(ul)	(void *)((unsigned long)ul)
#endif

#define osm_roundup(x, y) (		\
{					\
	const typeof(y) __y = y;	\
	(((x) + (__y - 1)) / __y) * __y;\
}					\
)
/*
************************************************************************
************************************************************************
*/
#if !defined(sg_page) && (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24))
	#define sg_page(sg)	sg->page
#endif
/*
************************************************************************
************************************************************************
*/
struct osm_list_head {
	struct osm_list_head *next, *prev;
};
#define osm_list_entry(ptr, type, member)				\
	container_of(ptr, type, member)

#define osm_list_first_entry(ptr, type, member)				\
	list_entry((ptr)->next, type, member)

#define osm_list_for_each(pos, head)					\
	for (pos = (head)->next; prefetch(pos->next), pos != (head);	\
		pos = pos->next)

#define __osm_list_for_each(pos, head)					\
	for (pos = (head)->next; pos != (head); pos = pos->next)

#define osm_list_for_each_entry(pos, head, member)			\
	for (pos = osm_list_entry((head)->next, typeof(*pos), member);	\
	     prefetch(pos->member.next), &pos->member != (head);	\
	     pos = osm_list_entry(pos->member.next, typeof(*pos), member))

#define osm_list_for_each_entry_safe(pos, n, head, member)		\
	for (pos = list_entry((head)->next, typeof(*pos), member),	\
		n = list_entry(pos->member.next, typeof(*pos), member);	\
		&pos->member != (head); 				\
		pos = n, n = list_entry(n->member.next, typeof(*n), member))
/*
************************************************************************
************************************************************************
*/
#define osm_msleep(x)		msleep(x)
#define osm_mdelay(x)		mdelay(x)
#define osm_udelay(x)		udelay(x)
#define osm_wait_after(x)	({	\
	unsigned long time;		\
	time = jiffies;			\
	while (time_after(jiffies, time + msecs_to_jiffies(x))) {	\
		break;			\
	}				\
})
/*
************************************************************************
************************************************************************
*/
#define osm_printk(format, arg...)	\
	do {				\
		mdelay(1);		\
		printk(format, ##arg);	\
		mdelay(3);		\
	} while (0)
#define _ASSERT()  do { osm_printk("ASSERT: function:%s, line:%d\n", __func__, __LINE__);} while(0)
#define OSM_INIT_WORK(_work, _func, _dev...)		INIT_WORK(_work, _func, ##_dev)
#define OSM_INIT_DELAYED_WORK(_work, _func, _dev...)	INIT_DELAYED_WORK(_work, _func, ##_dev)
/*
************************************************************************
************************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
	extern void arcsas_add_scsi_device_wqfn(struct work_struct *work);
	extern void arcsas_scsi_scan_host_wqfn(struct work_struct *work);
	extern void arcsas_do_smp_phy_control_reset_wqfn(struct work_struct *work);
	extern void arcsas_do_smp_state_machine_wqfn(struct work_struct *work);
	extern void arcsas_remove_scsi_device_wqfn(struct work_struct *work);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	extern void arcsas_add_scsi_device_wqfn(void *acb);
	extern void arcsas_scsi_scan_host_wqfn(void *acb);
	extern void arcsas_do_smp_phy_control_reset_wqfn(void *acb);
	extern void arcsas_do_smp_state_machine_wqfn(void *acb);
	extern void arcsas_remove_scsi_device_wqfn(void *acb);
#endif
extern unsigned long get_jiffies(void);
extern void osm_atomic_inc(atomic_t *v);	
extern int osm_atomic_inc_return(atomic_t *v);
extern int osm_atomic_read(const atomic_t *v);
extern void osm_atomic_dec(atomic_t *v);
extern void osm_atomic_set(atomic_t *addr, int v);
extern int osm_pci_read_config_dword(struct pci_dev *dev, int offset, u32 *ptr);
extern int osm_pci_read_config_word(struct pci_dev *dev, int offset, u16 *ptr);
extern int osm_pci_read_config_byte(struct pci_dev *dev, int offset, u8 *ptr);
extern int osm_pci_write_config_dword(struct pci_dev *dev, int offset, u32 val);	
extern int osm_pci_write_config_word(struct pci_dev *dev, int offset, u16 val);
extern int osm_pci_write_config_byte(struct pci_dev *dev, int offset, u8 val);
extern void osm_memset(void *buf, int  value, size_t len);
extern void osm_memcpy(void *dest, void *source, size_t len);
extern int osm_memcmp(void *buf0, void *buf1, size_t len);
extern int osm_strncmp(const char *cs, const char *ct, size_t count);
extern int osm_schedule_work(struct work_struct *work);			
extern void arcsas_spin_lock_init(spinlock_t *lock);
extern void arcsas_spin_lock_irqsave(spinlock_t *lock, unsigned long *flags);
extern void arcsas_spin_unlock_irqrestore(spinlock_t *lock, unsigned long *flags);
extern void arcsas_spin_lock(spinlock_t *lock);
extern void arcsas_spin_unlock(spinlock_t *lock);
extern void swap_buf_le16(u16 *buf, unsigned int buf_words);
extern struct scsi_device *__osm_scsi_device_lookup(struct Scsi_Host *shost, uint channel, uint id, uint lun);
extern struct scsi_device *osm_scsi_device_lookup(struct Scsi_Host *shost, uint channel, uint id, uint lun);
extern void osm_scsi_device_put(struct scsi_device *sdev);
extern void osm_writel(u32 val, void *addr);
extern u32 osm_readl(void *addr);
extern void osm_outl(u32 val, unsigned long addr);
extern u32 osm_inl(unsigned long addr);
extern void arcsas_touch_watchdog(void);
extern u32 arcsas_show_time(void);

